/****************************************************************************/
/* Another multi-workflow scheduler                                         */
/* This scheduler is based on HEFT algorithm                                */
/*                                                                          */
/* Author(s):                                                               */
/* - Benjamin ISNARD (benjamin.isnard@ens-lyon.fr)                          */
/*                                                                          */
/*  This file is part of DIET .                                             */
/*                                                                          */
/*  Copyright ENS Lyon, INRIA, UCBL, SysFera (2000)                         */
/*                                                                          */
/*  - Frederic.Desprez@ens-lyon.fr (Project Manager)                        */
/*  - Eddy.Caron@ens-lyon.fr (Technical Manager)                            */
/*  - Tech@sysfera.com (Maintainer and Technical Support)                   */
/*                                                                          */
/*  This software is a computer program whose purpose is to provide an      */
/*  easy and transparent access to distributed and heterogeneous            */
/*  platforms.                                                              */
/*                                                                          */
/*                                                                          */
/*  This software is governed by the CeCILL license under French law and    */
/*  abiding by the rules of distribution of free software.  You can  use,   */
/*  modify and/ or redistribute the software under the terms of the CeCILL  */
/*  license as circulated by CEA, CNRS and INRIA at the following URL       */
/*  "http://www.cecill.info".                                               */
/*                                                                          */
/*  As a counterpart to the access to the source code and  rights to copy,  */
/*  modify and redistribute granted by the license, users are provided      */
/*  only with a limited warranty  and the software's author,  the holder    */
/*  of the economic rights,  and the successive licensors  have only        */
/*  limited liability.                                                      */
/*                                                                          */
/*  In this respect, the user's attention is drawn to the risks             */
/*  associated with loading,  using,  modifying and/or developing or        */
/*  reproducing the software by the user in light of its specific status    */
/*  of free software, that may mean  that it is complicated to              */
/*  manipulate, and  that  also therefore means  that it is reserved for    */
/*  developers and experienced professionals having in-depth computer       */
/*  knowledge. Users are therefore encouraged to load and test the          */
/*  software's suitability as regards their requirements in conditions      */
/*  enabling the security of their systems and/or data to be ensured and,   */
/*  more generally, to use and operate it in the same conditions as         */
/*  regards security.                                                       */
/*                                                                          */
/*  The fact that you are presently reading this means that you have had    */
/*  knowledge of the CeCILL license and that you accept its terms.          */
/*                                                                          */
/****************************************************************************/

/* $Id$
 * $Log$
 * Revision 1.10  2010/03/31 21:15:39  bdepardo
 * Changed C headers into C++ headers
 *
 * Revision 1.9  2009/02/06 14:50:35  bisnard
 * setup exceptions
 *
 * Revision 1.8  2009/01/16 13:41:22  bisnard
 * added common base class DagScheduler to simplify dag events handling
 * improved exception management
 *
 * Revision 1.7  2008/10/14 13:24:49  bisnard
 * use new class structure for dags (DagNode,DagNodePort)
 *
 * Revision 1.6  2008/07/24 21:08:11  rbolze
 * New multi-wf heuristic FCFS (First Come First Serve)
 *
 * Revision 1.5  2008/06/25 10:05:44  bisnard
 * - Waiting priority set when node is put back in waiting queue
 * - Node index in wf_response stored in Node class (new attribute submitIndex)
 * - HEFT scheduler uses SeD ref instead of hostname
 * - Estimation vector and ReqID passed to client when SeD chosen by MaDag
 * - New params in execNodeOnSeD to provide ReqId and estimation vector
 * to client for solve request
 *
 * Revision 1.4  2008/06/19 10:18:54  bisnard
 * new heuristic AgingHEFT for multi-workflow scheduling
 *
 * Revision 1.3  2008/06/18 15:06:51  bisnard
 * set NodeRun class as friend to protect handler methods
 *
 * Revision 1.2  2008/06/01 14:06:57  rbolze
 * replace most ot the cout by adapted function from debug.cc
 * there are some left ...
 *
 * Revision 1.1  2008/04/30 07:37:01  bisnard
 * use relative timestamps for estimated and real completion time
 * make MultiWfScheduler abstract and add HEFT MultiWf scheduler
 *
 */

#ifndef _MULTIWFHEFT_HH_
#define _MULTIWFHEFT_HH_

#include <cmath>

#include "debug.hh"

#include "MultiWfScheduler.hh"


using namespace std;

namespace madag {

/****************************************************************************/
/*                              Multi-HEFT                                  */
/****************************************************************************/

  class MultiWfHEFT : public MultiWfScheduler {

  public:

    MultiWfHEFT(MaDag_impl* maDag);
    virtual ~MultiWfHEFT();

  protected:
    /**
     * Updates scheduler when a node has been executed
     */
    virtual void
        handlerNodeDone(DagNode * node);

  }; // end class MultiWfHEFT

/****************************************************************************/
/*                              Aging HEFT                                  */
/****************************************************************************/

  class MultiWfAgingHEFT : public MultiWfScheduler {

      public:

    MultiWfAgingHEFT(MaDag_impl* maDag);
    virtual ~MultiWfAgingHEFT();

  protected:

    /**
     * internal dag scheduling
     */
    virtual void
        intraDagSchedule(Dag * dag, MasterAgent_var MA)
        throw (MaDag::ServiceNotFound, MaDag::CommProblem);

    /**
     * Updates scheduler when a node has been executed
     */
    virtual void
        handlerNodeDone(DagNode * node);

    /**
     * set node priority before inserting into execution queue
     * (called by run method)
     * @param node   the node to insert
     */
    virtual void
        setExecPriority(DagNode * node);

    /**
     * set node priority before inserting back in the ready queue
     */
    virtual void
        setWaitingPriority(DagNode * node);

    /**
     * Save the state of dags
     */
    map<Dag*, DagState> dagsState;

  }; // end class MultiWfAgingHEFT

} // end namespace madag

#endif // _MULTIWFHEFT_HH_
