/****************************************************************************/
/* Interface for limiting access to a resource to a configurable number of  */
/* consumers.                                                               */
/*                                                                          */
/*  Author(s):                                                              */
/*    - Holly DAIL (Holly.Dail@ens-lyon.fr)                                 */
/*                                                                          */
/*  This file is part of DIET .                                             */
/*                                                                          */
/*  Copyright ENS Lyon, INRIA, UCBL, SysFera (2000)                         */
/*                                                                          */
/*  - Frederic.Desprez@ens-lyon.fr (Project Manager)                        */
/*  - Eddy.Caron@ens-lyon.fr (Technical Manager)                            */
/*  - Tech@sysfera.com (Maintainer and Technical Support)                   */
/*                                                                          */
/*  This software is a computer program whose purpose is to provide an      */
/*  easy and transparent access to distributed and heterogeneous            */
/*  platforms.                                                              */
/*                                                                          */
/*                                                                          */
/*  This software is governed by the CeCILL license under French law and    */
/*  abiding by the rules of distribution of free software.  You can  use,   */
/*  modify and/ or redistribute the software under the terms of the CeCILL  */
/*  license as circulated by CEA, CNRS and INRIA at the following URL       */
/*  "http://www.cecill.info".                                               */
/*                                                                          */
/*  As a counterpart to the access to the source code and  rights to copy,  */
/*  modify and redistribute granted by the license, users are provided      */
/*  only with a limited warranty  and the software's author,  the holder    */
/*  of the economic rights,  and the successive licensors  have only        */
/*  limited liability.                                                      */
/*                                                                          */
/*  In this respect, the user's attention is drawn to the risks             */
/*  associated with loading,  using,  modifying and/or developing or        */
/*  reproducing the software by the user in light of its specific status    */
/*  of free software, that may mean  that it is complicated to              */
/*  manipulate, and  that  also therefore means  that it is reserved for    */
/*  developers and experienced professionals having in-depth computer       */
/*  knowledge. Users are therefore encouraged to load and test the          */
/*  software's suitability as regards their requirements in conditions      */
/*  enabling the security of their systems and/or data to be ensured and,   */
/*  more generally, to use and operate it in the same conditions as         */
/*  regards security.                                                       */
/*                                                                          */
/*  The fact that you are presently reading this means that you have had    */
/*  knowledge of the CeCILL license and that you accept its terms.          */
/*                                                                          */
/****************************************************************************/
/* $Id$
 * $Log$
 * Revision 1.5  2010/03/31 21:15:40  bdepardo
 * Changed C headers into C++ headers
 *
 * Revision 1.4  2007/04/16 22:43:44  ycaniou
 * Make all necessary changes to have the new option HAVE_ALT_BATCH operational.
 * This is indented to replace HAVE_BATCH.
 *
 * First draw to manage batch systems with a new Cori plug-in.
 *
 * Revision 1.3  2005/04/25 09:08:12  hdail
 * Print warning if jobs exit SeD-level queue out of order, but don't re-insert
 * them.  This approach needs to be redone later.
 *
 * Revision 1.2  2004/12/08 15:02:52  alsu
 * plugin scheduler first-pass validation testing complete.  merging into
 * main CVS trunk; ready for more rigorous testing.
 *
 * Revision 1.1.2.1  2004/12/01 14:53:44  alsu
 * removing spurious argument that caused compilation warning (and thus
 * failure in maintainer mode) on certain "sensitive" versions of gcc.
 *
 * Revision 1.1  2004/10/04 13:55:06  hdail
 * - Added AccessController class, an enhanced counting semaphore.
 * - Added config file options for controlling concurrent SeD access.
 *
 ****************************************************************************/

#include "AccessController.hh"

#include <cstdio>
#include <iostream>
using namespace std;
#include <cstdlib>

#include "omnithread.h"
#include "debug.hh"

/** The trace level. */
extern unsigned int TRACE_LEVEL;

#define ACCESS_TRACE_FUNCTION(formatted_text) \
  TRACE_TEXT(TRACE_ALL_STEPS, "AccessControl::"); \
  TRACE_FUNCTION(TRACE_ALL_STEPS,formatted_text)

AccessController::AccessController(int initialResources) {
  if(initialResources >= 0) {
    this->resourcesInitial = initialResources;
  } else {
    this->resourcesInitial = 0;
    fprintf(stderr, "AccessController:: invalid initialResources of %d.\n",
        initialResources);
    fprintf(stderr, "AccessController:: setting initialResources to 0.\n");
  } 

  resourceSemaphore = new omni_semaphore(this->resourcesInitial);
  numFreeSlots = this->resourcesInitial;
  numWaiting = 0;

  reqCounter = 0;
  maxIDReleased = -1;
}

void
AccessController::waitForResource(){
  int myReqID = -1;
    
  this->globalLock.lock();      /** LOCK */
  myReqID = this->reqCounter++;
  this->numWaiting++;
  this->globalLock.unlock();    /** UNLOCK */

  TRACE_TEXT(TRACE_ALL_STEPS, "Thread " <<  (omni_thread::self())->id()
        << " / Request " << myReqID << " enqueued ("
        << this->numWaiting << " waiting, "
        << this->numFreeSlots << " slots free)" << endl);

  this->resourceSemaphore->wait();

  // TODO: On systems tested semaphore gives FIFO ordering and the following
  // test succeeds only in the case that tasks switch order between the receipt
  // of request IDs and the wait on the semaphore.  This case is not very
  // important because those tasks arrive at almost the same moment.
  // However, semaphores do not guarantee FIFO ordering so for portability
  // another solution must be found to guarantee FIFO.
  if( myReqID != (this->maxIDReleased + 1) ) {
    WARNING("Thread " << (omni_thread::self())->id()
          << " / Request " << myReqID << " exiting queue out-of-order.");
  }

  if(this->numFreeSlots <= 0){
    fprintf(stderr, 
            "AccessController:: confusion between "
            "semaphore and numFreeSlots ...");
  }
  if(this->numWaiting <= 0){
    fprintf(stderr, 
            "AccessController:: Unexplained problem "
            "counting waiting threads.");
  }

  this->globalLock.lock();      /** LOCK */
  this->numFreeSlots--;
  this->numWaiting--;
  if(myReqID > this->maxIDReleased){
    maxIDReleased = myReqID;
  }
  this->globalLock.unlock();    /** UNLOCK */

  TRACE_TEXT(TRACE_ALL_STEPS, "Thread " <<  (omni_thread::self())->id()
      << " / Request " << myReqID << " got resource ("
      << this->numWaiting << " waiting, "
      << this->numFreeSlots << " slots free)" << endl);
}

void
AccessController::releaseResource(){
  this->globalLock.lock();      /** LOCK */
  this->numFreeSlots++;
  this->globalLock.unlock();    /** UNLOCK */  

  TRACE_TEXT(TRACE_ALL_STEPS, "Thread " <<  (omni_thread::self())->id()
      << " released resource ("
      << this->numWaiting << " waiting, "
      << this->numFreeSlots << " slots free)" << endl);

  this->resourceSemaphore->post();
}

int AccessController::getFreeSlotCount(){
  return this->numFreeSlots;
}

int AccessController::getNumWaiting(){
  return this->numWaiting;
}

int AccessController::getTotalResourceCount(){
  return this->resourcesInitial;
}

