/***********************************************************/
/* Class used to manage relationships between data items   */
/* (used for containers in DAGDA)                          */
/*                                                         */
/*  Author(s):                                             */
/*    - Benjamin Isnard (benjamin.isnard@ens-lyon.fr       */
/*                                                         */
/*  This file is part of DIET .                                             */
/*                                                                          */
/*  Copyright ENS Lyon, INRIA, UCBL, SysFera (2000)                         */
/*                                                                          */
/*  - Frederic.Desprez@ens-lyon.fr (Project Manager)                        */
/*  - Eddy.Caron@ens-lyon.fr (Technical Manager)                            */
/*  - Tech@sysfera.com (Maintainer and Technical Support)                   */
/*                                                                          */
/*  This software is a computer program whose purpose is to provide an      */
/*  easy and transparent access to distributed and heterogeneous            */
/*  platforms.                                                              */
/*                                                                          */
/*                                                                          */
/*  This software is governed by the CeCILL license under French law and    */
/*  abiding by the rules of distribution of free software.  You can  use,   */
/*  modify and/ or redistribute the software under the terms of the CeCILL  */
/*  license as circulated by CEA, CNRS and INRIA at the following URL       */
/*  "http://www.cecill.info".                                               */
/*                                                                          */
/*  As a counterpart to the access to the source code and  rights to copy,  */
/*  modify and redistribute granted by the license, users are provided      */
/*  only with a limited warranty  and the software's author,  the holder    */
/*  of the economic rights,  and the successive licensors  have only        */
/*  limited liability.                                                      */
/*                                                                          */
/*  In this respect, the user's attention is drawn to the risks             */
/*  associated with loading,  using,  modifying and/or developing or        */
/*  reproducing the software by the user in light of its specific status    */
/*  of free software, that may mean  that it is complicated to              */
/*  manipulate, and  that  also therefore means  that it is reserved for    */
/*  developers and experienced professionals having in-depth computer       */
/*  knowledge. Users are therefore encouraged to load and test the          */
/*  software's suitability as regards their requirements in conditions      */
/*  enabling the security of their systems and/or data to be ensured and,   */
/*  more generally, to use and operate it in the same conditions as         */
/*  regards security.                                                       */
/*                                                                          */
/*  The fact that you are presently reading this means that you have had    */
/*  knowledge of the CeCILL license and that you accept its terms.          */
/*                                                                          */
/***********************************************************/
/* $Id$
 * $Log$
 * Revision 1.6  2010/07/30 14:44:26  glemahec
 * Temporary corrections for the new compilation process. Parallel compilation is still broken and there is a big mess on the CMakeLists files...
 *
 * Revision 1.5  2009/10/02 07:43:21  bisnard
 * reduced trace verbosity for containers
 *
 * Revision 1.4  2009/09/25 12:51:31  bisnard
 * fixed bug in case of containers containing empty elements as first elements
 *
 * Revision 1.3  2009/03/27 08:06:20  bisnard
 * new method getRelationNb
 *
 * Revision 1.2  2009/01/16 13:36:54  bisnard
 * make thread-safe using mutex
 *
 * Revision 1.1  2008/09/09 10:07:36  bisnard
 * new class used for container mgmt using Dagda agent
 *
 *
 ***********************************************************/

#include "DataRelationMgr.hh"
#include "debug.hh"
#include <iostream>

DataRelationMgr::DataRelationMgr() { }

void
DataRelationMgr::addRelation(const string& dataID1,
                             const string& dataID2,
                             long index,
                             long flag) {
  dataRelationValue_t value = { dataID2, index, flag};
  myLock.lock(); /** LOCK */
  myMap.insert(make_pair(dataID1, value));
  myLock.unlock(); /** UNLOCK */
}

void
DataRelationMgr::remRelation(const string& dataID1,
                             long index) {
  myLock.lock(); /** LOCK */
  for (multimap<std::string, dataRelationValue_t>::iterator iter = myMap.lower_bound(dataID1);
       iter != myMap.upper_bound(dataID1);
       ++iter) {
    if (iter->second.index == index) {
      myMap.erase(iter);
      break;
    }
  }
  myLock.unlock(); /** UNLOCK */
}

void
DataRelationMgr::remAllRelation(const string& dataID, bool reverse) {
  // Delete links FROM the item
  myLock.lock();
  int count = myMap.erase(dataID);
  myLock.unlock();
  TRACE_TEXT(TRACE_ALL_STEPS, "Removed " << count << " relations for key "
      << dataID << endl);
  // Delete links TO the item
  if (reverse) {
    // TODO delete all links TO the item
  }
}

/**
 * Get the list of relations for a given dataID
 * Note: the SeqString and SeqLong must be pre-allocated by the caller
 */
void
DataRelationMgr::getRelationList(const string& dataID,
                                 SeqString& dataIDList,
                                 SeqLong& flagList,
                                 bool ordered) {
  int ix = 0;
  int listSize = dataIDList.length();
  myLock.lock();
  for (multimap<std::string, dataRelationValue_t>::iterator iter = myMap.lower_bound(dataID);
       iter != myMap.upper_bound(dataID);
       ++iter) {
    int jx = ordered ? iter->second.index : ix++;
    if (jx >= listSize) {
      INTERNAL_ERROR("Mismatching container element index with nb of relationships",1);
    }
    dataIDList[jx] = CORBA::string_dup(iter->second.ID.c_str());
    flagList[jx]   = iter->second.flag;
  }
  myLock.unlock();
}

unsigned int
DataRelationMgr::getRelationNb(const string& dataID) {
  unsigned int ix = 0;
  myLock.lock();
  for (multimap<std::string, dataRelationValue_t>::iterator iter = myMap.lower_bound(dataID);
       iter != myMap.upper_bound(dataID);
       ++iter) { ++ix; }
  myLock.unlock();
  return ix;
}

unsigned int
DataRelationMgr::getRelationMaxIndex(const string& dataID) {
  int max = 0;
  myLock.lock();
  for (multimap<std::string, dataRelationValue_t>::iterator iter = myMap.lower_bound(dataID);
       iter != myMap.upper_bound(dataID);
       ++iter) {
    if (iter->second.index > max)
      max = (unsigned int) iter->second.index;
  }
  myLock.unlock();
  return max;
}

/**
 * Display all the relations for debug
 */
void
DataRelationMgr::displayContent() {
  if (TRACE_LEVEL >= TRACE_ALL_STEPS) {
    cout << "Content of relationship DB: " << endl;
    myLock.lock();
    for (multimap<string, dataRelationValue_t>::iterator iter = myMap.begin();
         iter != myMap.end();
         ++iter) {
      cout << " | " << iter->first << " | " << iter->second.ID
           << " | " << iter->second.index << " | " << iter->second.flag << endl;
    }
    myLock.unlock();
  }
}
