/****************************************************************************/
/* The port classes used for dag execution                                  */
/* The main classes are DagNodePort, DagNodeOutPort and DagNodeInPort       */
/*                                                                          */
/* Author(s):                                                               */
/* - Benjamin ISNARD (benjamin.isnard@ens-lyon.fr)                          */
/*                                                                          */
/*  This file is part of DIET .                                             */
/*                                                                          */
/*  Copyright ENS Lyon, INRIA, UCBL, SysFera (2000)                         */
/*                                                                          */
/*  - Frederic.Desprez@ens-lyon.fr (Project Manager)                        */
/*  - Eddy.Caron@ens-lyon.fr (Technical Manager)                            */
/*  - Tech@sysfera.com (Maintainer and Technical Support)                   */
/*                                                                          */
/*  This software is a computer program whose purpose is to provide an      */
/*  easy and transparent access to distributed and heterogeneous            */
/*  platforms.                                                              */
/*                                                                          */
/*                                                                          */
/*  This software is governed by the CeCILL license under French law and    */
/*  abiding by the rules of distribution of free software.  You can  use,   */
/*  modify and/ or redistribute the software under the terms of the CeCILL  */
/*  license as circulated by CEA, CNRS and INRIA at the following URL       */
/*  "http://www.cecill.info".                                               */
/*                                                                          */
/*  As a counterpart to the access to the source code and  rights to copy,  */
/*  modify and redistribute granted by the license, users are provided      */
/*  only with a limited warranty  and the software's author,  the holder    */
/*  of the economic rights,  and the successive licensors  have only        */
/*  limited liability.                                                      */
/*                                                                          */
/*  In this respect, the user's attention is drawn to the risks             */
/*  associated with loading,  using,  modifying and/or developing or        */
/*  reproducing the software by the user in light of its specific status    */
/*  of free software, that may mean  that it is complicated to              */
/*  manipulate, and  that  also therefore means  that it is reserved for    */
/*  developers and experienced professionals having in-depth computer       */
/*  knowledge. Users are therefore encouraged to load and test the          */
/*  software's suitability as regards their requirements in conditions      */
/*  enabling the security of their systems and/or data to be ensured and,   */
/*  more generally, to use and operate it in the same conditions as         */
/*  regards security.                                                       */
/*                                                                          */
/*  The fact that you are presently reading this means that you have had    */
/*  knowledge of the CeCILL license and that you accept its terms.          */
/*                                                                          */
/****************************************************************************/
/* $Id$
 * $Log$
 * Revision 1.15  2010/07/20 09:20:11  bisnard
 * integration with eclipse gui and with dietForwarder
 *
 * Revision 1.14  2009/08/26 10:33:09  bisnard
 * implementation of workflow status & restart
 *
 * Revision 1.13  2009/07/30 09:43:09  bisnard
 * modified methods to free persistent data generated by workflow execution
 *
 * Revision 1.12  2009/07/07 09:02:00  bisnard
 * modified toXML method
 *
 * Revision 1.11  2009/06/15 12:24:29  bisnard
 * new class DagNodeArgPort (arg ports not used for funct wf anymore)
 * use WfDataWriter class to display data
 *
 * Revision 1.10  2009/05/15 11:15:07  bisnard
 * container initialization adapted to standard notation
 * added method to retrieve container cardinal
 *
 * Revision 1.9  2009/04/17 09:02:15  bisnard
 * container empty elements (added WfVoidAdapter class)
 *
 * Revision 1.8  2009/02/06 14:53:09  bisnard
 * - setup exceptions
 * - display methods for container results
 *
 * Revision 1.7  2009/01/16 13:55:36  bisnard
 * changes in dag event handling methods
 *
 * Revision 1.6  2008/12/09 12:15:59  bisnard
 * pending instanciation handling (uses dag outputs for instanciation
 * of a functional wf)
 *
 * Revision 1.5  2008/12/02 10:14:51  bisnard
 * modified nodes links mgmt to handle inter-dags links
 *
 * Revision 1.4  2008/10/30 14:33:01  bisnard
 * added recursive container initialization
 *
 * Revision 1.3  2008/10/22 09:29:00  bisnard
 * replaced uint by standard type
 *
 * Revision 1.2  2008/10/20 07:59:29  bisnard
 * file header creation
 *
 */

#ifndef _DAGNODEPORT_HH_
#define _DAGNODEPORT_HH_

#include <string>
#include <list>

#include "DIET_data.h"
#include "WfPortAdapter.hh"
#include "WfUtils.hh"
#include "WfPort.hh"
#include "DagNode.hh"

using namespace std;

/*****************************************************************************/
/*                        DagNodePort (ABSTRACT)                             */
/*****************************************************************************/

class DagNodePort : public WfPort {

public:
  
  /**
   * Separator used for initializing or displaying container elements
   */
  static string containerSeparator;
  
  /**
   * Event Message types
   */
  enum eventMsg_e {
    DATAID
  };
  
  /***************************************************/
  /* constructors/destructor                         */
  /***************************************************/

  DagNodePort(DagNode * parent,
              string _id,
              WfPort::WfPortType _portType,
              WfCst::WfDataType _type,
              unsigned int _depth,
              unsigned int _ind) ;
  
  /***************************************************/
  /*               public methods                    */
  /***************************************************/

  /**
   * Initialize the profile before node submission
   */
  virtual void
  initProfileSubmit();

  /**
   * Initialize the profile before node execution
   */
  virtual void
  initProfileExec() throw (WfDataException) = 0;
  
  /**
   * Return the object description as a string
   */
  virtual string
  toString() const;

  /**
   * Return the XML description of the port
   */
  virtual void
  toXML(ostream& output) const;

  /**
   * Return the profile of the node
   * used by WfPortAdapter::getSourceDataID
   */
  diet_profile_t *
  profile();
  
  /**
   * Set the data ID of the data received/produced by the port
   */
  void
  setDataID(const string& dataID);

  /**
   * Return the data ID of the port
   * @exception WfDataException(eID_UNDEF) if the data ID is not defined
   */
  const string&
  getDataID() throw (WfDataException);
  
  bool
  isDataIDAvailable(MasterAgent_var& MA);

  /**
   * Free the persistent data
   */
  virtual void
  freePersistentData(MasterAgent_var& MA) = 0;

protected:

  /**
   * Returns the persistence mode for this port
   */
  virtual diet_persistence_mode_t
      getPersistenceMode() = 0;

  /**
   * Initializes the profile when no value is provided
   */
  void
  setProfileWithoutValue();

  /**
   * The port data ID
   */
  string myDataID;

  /**
   * The reference of port node
   */
  DagNode * myParent;

}; // end DagNodePort


/*****************************************************************************/
/*                           DagNodeOutPort                                  */
/*****************************************************************************/

class DagNodeOutPort : virtual public DagNodePort {
public:
  DagNodeOutPort(DagNode * parent,
                 string _id,
                 WfCst::WfDataType _type,
                 unsigned int _depth,
                 unsigned int _ind);

  ~DagNodeOutPort();

  /**
   * Initialize the profile before node execution
   */
  virtual void
  initProfileExec() throw (WfDataException);
  
  /**
   * Store the data IDs from the profile
   */
  void
  storeProfileData();

  /**
   * Free the persistent data
   */
  void
  freePersistentData(MasterAgent_var& MA);
  
  /**
   * Return the object description as a string
   */
  virtual string
  toString() const;

  /**
   * Return the XML description of the port
   */
  virtual void
  toXML(ostream& output) const;

  /**
   * Get the cardinal of the content of the output data (when it's a container)
   * (for any ID within the container tree of the output)
   * @param dataID  the ID of the data container
   * @return  the size of the container
   */
//   unsigned int
//   getDataIDCardinal(const string& dataID) throw (WfDataException);

  /**
   * Get the ID of an element of the output data (when it's a container)
   * @param eltIdx a list of indexes for the element
   * @return  the data ID (may be empty if the element is NULL)
   * @exception WfDataException(eID_UNDEF) if the data ID is not defined
   */
  string
  getElementDataID(const list<unsigned int>& eltIdx) throw (WfDataException);

  /**
   * Write the output data in a stream
   * If the data is a container then it's displayed as a parenthezized
   * comma-separated list
   * @param dataWriter  the data output formatter
   */
  void
  writeData(WfDataWriter* dataWriter) throw (WfDataException);

  /**
   * Write an element of the output data (when it's a container)
   * @param dataWriter  the data output formatter
   * @param eltIdx a list of indexes for the element
   */
  void
  writeDataElement(WfDataWriter* dataWriter, const list<unsigned int>& idxList)
      throw (WfDataException);


protected:

  /**
   * Display method used only when port type is container
   * (recursive)
   */
  void
  writeContainer(WfDataWriter* dataWriter,
                 const string& containerID,
                 unsigned int depth) throw (WfDataException);
  /**
   * Display method used for container elements (dagda API)
   */
  void
  writeContainerData(WfDataWriter* dataWriter,
                     const string& eltID);

  /**
   * Returns the persistence mode for this port
   */
  diet_persistence_mode_t
      getPersistenceMode();

};


/*****************************************************************************/
/*                           DagNodeInPort                                   */
/*****************************************************************************/

class DagNodeInPort : virtual public DagNodePort {
public:
  DagNodeInPort(DagNode * parent,
                string _id,
                WfCst::WfDataType _type,
                unsigned int _depth,
                unsigned int _ind);
  
  /**
   * Event Message types
   */
  enum eventMsg_e {
    DEPEND
  };
  
  /**
   * Set the source of the input port (when dag built by parsing XML)
   * @param strRef The source port (output port) reference
   */
  virtual void
  setConnectionRef(const string& strRef);

  /**
   * Set the source of the input port (when dag built by wf instanciator)
   * @param strRef The source port (output port) reference
   */
  virtual void
  setPortAdapter(WfPortAdapter* adapter);
  
  /**
   * Initialize the profile before node execution
   */
  virtual void
  initProfileExec() throw (WfDataException);
  
  /**
   * Return the object description as a string
   */
  virtual string
  toString() const;

  /**
   * Return the XML description of the port
   */
  virtual void
  toXML(ostream& output) const;

  /**
   * Free the persistent data
   */
  void
  freePersistentData(MasterAgent_var& MA);

protected:
  /**
   * Source Data ID retrieval (used for node execution)
   * Initialize the profile corresponding to this port with the data
   * provided by the source port (uses data IDs because all output
   * data in a workflow is persistent)
   * @return false if source data cannot be found
   */
  virtual void
  initSourceData() throw (WfDataException) ;

  /**
   * Returns the persistence mode for this port
   */
  diet_persistence_mode_t
  getPersistenceMode();

}; // end DagNodeInPort


/*****************************************************************************/
/*                         DagNodeInOutPort                                  */
/*****************************************************************************/

class DagNodeInOutPort : virtual public DagNodeInPort,
                           virtual public DagNodeOutPort {
public:
  DagNodeInOutPort(DagNode * parent,
                   string _id,
                   WfCst::WfDataType _type,
                   unsigned int _depth,
                   unsigned int _ind);
  /**
   * Initialize the profile before node execution
   */
  virtual void
  initProfileExec() throw (WfDataException);
  
  /**
   * Return the object description as a string
   */
  virtual string
  toString() const;

  /**
   * Return the XML description of the port
   */
  virtual void
  toXML(ostream& output) const;

  /**
   * Free the persistent data
   */
  void
  freePersistentData(MasterAgent_var& MA);

protected:
  /**
   * Returns the persistence mode for this port
   */
  diet_persistence_mode_t
      getPersistenceMode();

}; // end DagNodeInOutPort


/*****************************************************************************/
/*                           DagNodeArgPort                                  */
/*****************************************************************************/

class DagNodeArgPort : virtual public DagNodePort {
public:
  DagNodeArgPort(DagNode * parent,
                 string _id,
                 WfCst::WfDataType _type,
                 unsigned int _depth,
                 unsigned int _ind);

  /**
   * Set the value of the data for the port
   */
  void
  setValue(const string& val);

  /**
   * Initialize the profile before node submission
   */
  virtual void
  initProfileSubmit();

  /**
   * Initialize the profile before node execution
   */
  virtual void
  initProfileExec() throw (WfDataException);
  
  /**
   * Return the object description as a string
   */
  virtual string
  toString() const;

  /**
   * Return the XML description of the port
   */
  virtual void
  toXML(ostream& output) const;

  /**
   * Free the persistent data
   */
  virtual void
  freePersistentData(MasterAgent_var& MA);

protected:

  /**
   * Initializes the profile when a value is provided
   */
  void
  setProfileWithValue();

  /**
   * Converts the string value to an allocated buffer for a matrix
   */
  void
  initMatrixValue(void **buffer);

  /**
   * Converts the string value to a data handle for a container
   * Examples of input: (3;4;5)  or ((toto;titi);(tata))
   */
  void
  initContainerValue();

  /**
   * Container initialization from string (recursive)
   * @param contIDPtr will contain the ID of the created container
   * @param contStr   string to parse (already parsed content is removed)
   * @param contDepth depth of the current container (must be >= 1)
   */
  void
  initContainerValueRec(char** contIDPtr,
                        string& contStr,
                        unsigned int contDepth);
  /**
   * Stores a data defined by a value string into the data manager
   * @param valStr  the value
   * @param valID   will contain the ID of the data
   */
  void
  initContainerElementValue(const char* valStr, char** valIDPtr);

  /**
   * Returns the persistence mode for this port
   */
  diet_persistence_mode_t
  getPersistenceMode();

  /**
   * The value of the data
   */
  string value;

}; // end DagNodeArgPort


#endif // _DAGNODEPORT_HH_
