/****************************************************************************/
/* The class representing a functional workflow                             */
/* Contains the workflow nodes and manages their instanciation as one or    */
/* several dags.                                                            */
/*                                                                          */
/* Author(s):                                                               */
/* - Benjamin ISNARD (benjamin.isnard@ens-lyon.fr)                          */
/*                                                                          */
/*  This file is part of DIET .                                             */
/*                                                                          */
/*  Copyright ENS Lyon, INRIA, UCBL, SysFera (2000)                         */
/*                                                                          */
/*  - Frederic.Desprez@ens-lyon.fr (Project Manager)                        */
/*  - Eddy.Caron@ens-lyon.fr (Technical Manager)                            */
/*  - Tech@sysfera.com (Maintainer and Technical Support)                   */
/*                                                                          */
/*  This software is a computer program whose purpose is to provide an      */
/*  easy and transparent access to distributed and heterogeneous            */
/*  platforms.                                                              */
/*                                                                          */
/*                                                                          */
/*  This software is governed by the CeCILL license under French law and    */
/*  abiding by the rules of distribution of free software.  You can  use,   */
/*  modify and/ or redistribute the software under the terms of the CeCILL  */
/*  license as circulated by CEA, CNRS and INRIA at the following URL       */
/*  "http://www.cecill.info".                                               */
/*                                                                          */
/*  As a counterpart to the access to the source code and  rights to copy,  */
/*  modify and redistribute granted by the license, users are provided      */
/*  only with a limited warranty  and the software's author,  the holder    */
/*  of the economic rights,  and the successive licensors  have only        */
/*  limited liability.                                                      */
/*                                                                          */
/*  In this respect, the user's attention is drawn to the risks             */
/*  associated with loading,  using,  modifying and/or developing or        */
/*  reproducing the software by the user in light of its specific status    */
/*  of free software, that may mean  that it is complicated to              */
/*  manipulate, and  that  also therefore means  that it is reserved for    */
/*  developers and experienced professionals having in-depth computer       */
/*  knowledge. Users are therefore encouraged to load and test the          */
/*  software's suitability as regards their requirements in conditions      */
/*  enabling the security of their systems and/or data to be ensured and,   */
/*  more generally, to use and operate it in the same conditions as         */
/*  regards security.                                                       */
/*                                                                          */
/*  The fact that you are presently reading this means that you have had    */
/*  knowledge of the CeCILL license and that you accept its terms.          */
/*                                                                          */
/****************************************************************************/
/* $Id$
 * $Log$
 * Revision 1.26  2011/03/25 17:45:17  hguemar
 * fix build in src/utils/workflow/FWorkflow.cc
 *
 * Revision 1.25  2011/03/25 17:24:02  hguemar
 * fix cppcheck raised issues: stlSize again
 *
 * Revision 1.24  2011/02/04 14:29:54  bdepardo
 * Remove unused variables
 *
 * Revision 1.23  2010/07/20 09:20:11  bisnard
 * integration with eclipse gui and with dietForwarder
 *
 * Revision 1.22  2009/10/23 14:01:17  bisnard
 * improved exception handling
 * avoid double instanciation of sinks
 *
 * Revision 1.21  2009/09/25 12:39:18  bisnard
 * modified includes to reduce inter-dependencies
 *
 * Revision 1.20  2009/08/26 10:33:09  bisnard
 * implementation of workflow status & restart
 *
 * Revision 1.19  2009/07/30 09:43:09  bisnard
 * modified methods to free persistent data generated by workflow execution
 *
 * Revision 1.18  2009/07/23 12:30:10  bisnard
 * new method finalize() for functional wf nodes
 * removed const on currDataLine parameter for instance creation
 *
 * Revision 1.17  2009/07/10 12:55:59  bisnard
 * implemented while loop workflow node
 *
 * Revision 1.16  2009/07/07 09:03:22  bisnard
 * changes for sub-workflows (FWorkflow class now inherits from FProcNode)
 *
 * Revision 1.15  2009/06/23 12:23:58  bisnard
 * removed dependencies (includes)
 *
 * Revision 1.14  2009/06/15 12:11:13  bisnard
 * use new XML Parser (SAX) for data source file
 * use new class WfValueAdapter to avoid data duplication
 * use new method FNodeOutPort::storeData
 * changed method to compute total nb of data items
 *
 * Revision 1.13  2009/05/27 08:49:43  bisnard
 * - modified condition output: new IF_THEN and IF_ELSE port types
 * - implemented MERGE and FILTER workflow nodes
 *
 * Revision 1.12  2009/05/15 11:10:20  bisnard
 * release for workflow conditional structure (if)
 *
 * Revision 1.11  2009/04/17 08:54:43  bisnard
 * renamed Node class as WfNode
 *
 * Revision 1.10  2009/04/09 09:56:20  bisnard
 * refactoring due to new class FActivityNode
 *
 * Revision 1.9  2009/04/08 09:34:56  bisnard
 * pending nodes mgmt moved to FWorkflow class
 * FWorkflow and FNode state graph revisited
 * FNodePort instanciation refactoring
 *
 * Revision 1.8  2009/02/06 14:55:08  bisnard
 * setup exceptions
 *
 * Revision 1.7  2009/01/16 16:31:54  bisnard
 * added option to specify data source file name
 *
 * Revision 1.6  2009/01/16 13:54:50  bisnard
 * new version of the dag instanciator with iteration strategies for nodes with multiple input ports
 *
 * Revision 1.5  2008/12/09 12:15:59  bisnard
 * pending instanciation handling (uses dag outputs for instanciation
 * of a functional wf)
 *
 * Revision 1.4  2008/12/02 14:17:49  bisnard
 * manage multi-dag cancellation when one dag fails
 *
 * Revision 1.3  2008/12/02 10:09:36  bisnard
 * added instanciation methods
 *
 * Revision 1.2  2008/10/22 09:29:00  bisnard
 * replaced uint by standard type
 *
 * Revision 1.1  2008/10/20 08:09:58  bisnard
 * new classes for functional workflows
 *
 */
//#include <boost/shared_ptr.hpp>
#include <map>
#include <functional>

#include "debug.hh"
#include "Dag.hh"
#include "DagWfParser.hh"
#include "FWorkflow.hh"
#include "FActivityNode.hh"
#include "FIfNode.hh"
#include "FLoopNode.hh"
#include "EventTypes.hh"

using namespace events;

FWorkflow::FWorkflow(string id, string name): FProcNode(NULL, id), myName(name)
{
}


FWorkflow::FWorkflow(string id, string name, FWorkflow*  parentWf)
  : FProcNode(parentWf, id), myName(name)
{
}

FWorkflow::~FWorkflow() {
  // free nodes
  while (! myInterface.empty() ) {
    FNode * p = myInterface.begin()->second;
    myInterface.erase( myInterface.begin() );
    delete p;
  }
  while (! myProc.empty() ) {
    FProcNode * p = myProc.begin()->second;
    myProc.erase( myProc.begin() );
    delete p;
  }
}

string FWorkflow::getName() const
{
  return myName;
}

FWorkflow*
FWorkflow::getRootWorkflow() const
{
  if (getWorkflow() == NULL) return const_cast<FWorkflow*>(this);
  return getWorkflow()->getRootWorkflow();
}

WfNode *
FWorkflow::getNode(const string& nodeId) throw (WfStructException) {
  WfNode * node;
  try {
    node = (WfNode *) getProcNode(nodeId);
  } catch (WfStructException& e) {
    node = (WfNode *) getInterfaceNode(nodeId);
  }
  return node;
}

FProcNode *
FWorkflow::getProcNode(const string& id) throw (WfStructException) {
  FProcNode *node = NULL;
  myLock.lock();    /** LOCK */
  map<string, FProcNode*>::iterator p = this->myProc.find(id);
  if ( p != this->myProc.end())
    node = p->second;
  myLock.unlock();  /** UNLOCK */
  if (node == NULL)
    throw WfStructException(WfStructException::eUNKNOWN_NODE,"FNode id="+id);
  return node;
}

FNode *
FWorkflow::getInterfaceNode(const string& id) throw (WfStructException) {
  FNode *node = NULL;
  myLock.lock();    /** LOCK */
  map<string, FNode*>::iterator p = this->myInterface.find(id);
  if ( p != this->myInterface.end())
    node = p->second;
  myLock.unlock();  /** UNLOCK */
  if (node == NULL)
    throw WfStructException(WfStructException::eUNKNOWN_NODE,"FNode id="+id);
  return node;
}

void
FWorkflow::checkPrec(NodeSet* contextNodeSet) throw (WfStructException) {
  TRACE_TEXT(TRACE_ALL_STEPS, traceId() << "CHECKING WF PRECEDENCE START" << endl);
  // processor nodes
  for (map<string, FProcNode * >::iterator p = myProc.begin();
       p != myProc.end();
       ++p ) {
    FProcNode * node = (FProcNode*) p->second;
    node->setNodePrecedence(contextNodeSet);
    // sub-workflows
    FWorkflow* subWf = dynamic_cast<FWorkflow*>(node);
    if (subWf) {
      subWf->checkPrec(subWf);
    }
  }
  // interface nodes (does sth only for sink nodes)
  for (map<string, FNode * >::iterator p = myInterface.begin();
       p != myInterface.end();
       ++p ) {
    FNode * node = (FNode *) p->second;
    node->setNodePrecedence(contextNodeSet);
  }
  // ports of workflow (for sub-workflows only)
  for (map<string,WfPort*>::iterator portIter = ports.begin();
       portIter != ports.end();
       ++portIter) {
    FNodePort* wfPort = dynamic_cast<FNodePort*>((WfPort*) portIter->second);
    try {
      FNode* interfNode = getInterfaceNode(wfPort->getInterfaceRef());
      interfNode->connectToWfPort(wfPort);
    } catch (const WfStructException& e) {
      throw WfStructException(e.Type(), string("Unknown reference for sub-workflow interface : ")
                                        + e.Info());
    }
  }
  TRACE_TEXT(TRACE_ALL_STEPS, traceId() << "CHECKING WF PRECEDENCE END" << endl);
}

FActivityNode*
FWorkflow::createActivity(const string& id) throw (WfStructException)
{
  // check if node does not already exist
  try {
    getProcNode(id);
  } catch (WfStructException& e) {
    TRACE_TEXT (TRACE_ALL_STEPS,"Creating activity node : " << id << endl);
    FActivityNode* node = new FActivityNode(this, id);
    myProc[id] = node;
    return node;
  }
  throw WfStructException(WfStructException::eDUPLICATE_NODE,"FNode id="+id);
}

FIfNode*
FWorkflow::createIf(const string& id) throw (WfStructException)
{
  // check if node does not already exist
  try {
    getProcNode(id);
  } catch (WfStructException& e) {
    TRACE_TEXT (TRACE_ALL_STEPS,"Creating IF node : " << id << endl);
    FIfNode* node = new FIfNode(this, id);
    myProc[id] = node;
    return node;
  }
  throw WfStructException(WfStructException::eDUPLICATE_NODE,"FNode id="+id);
}

FMergeNode*
FWorkflow::createMerge(const string& id) throw (WfStructException)
{
  // check if node does not already exist
  try {
    getProcNode(id);
  } catch (WfStructException& e) {
    TRACE_TEXT (TRACE_ALL_STEPS,"Creating MERGE node : " << id << endl);
    FMergeNode* node = new FMergeNode(this, id);
    myProc[id] = node;
    return node;
  }
  throw WfStructException(WfStructException::eDUPLICATE_NODE,"FNode id="+id);
}

FFilterNode*
FWorkflow::createFilter(const string& id) throw (WfStructException)
{
  // check if node does not already exist
  try {
    getProcNode(id);
  } catch (WfStructException& e) {
    TRACE_TEXT (TRACE_ALL_STEPS,"Creating FILTER node : " << id << endl);
    FFilterNode* node = new FFilterNode(this, id);
    myProc[id] = node;
    return node;
  }
  throw WfStructException(WfStructException::eDUPLICATE_NODE,"FNode id="+id);
}

FLoopNode*
FWorkflow::createLoop(const string& id) throw (WfStructException)
{
  // check if node does not already exist
  try {
    getProcNode(id);
  } catch (WfStructException& e) {
    TRACE_TEXT (TRACE_ALL_STEPS,"Creating LOOP node : " << id << endl);
    FLoopNode* node = new FLoopNode(this, id);
    myProc[id] = node;
    return node;
  }
  throw WfStructException(WfStructException::eDUPLICATE_NODE,"FNode id="+id);
}

FWorkflow*
FWorkflow::createSubWorkflow(const string& id, const string& name) throw (WfStructException)
{
  try {
    getProcNode(id);
  } catch (WfStructException& e) {
    TRACE_TEXT (TRACE_ALL_STEPS,"Creating sub-workflow node : " << id << endl);
    FWorkflow* node = new FWorkflow(id, name, this);
    EventManager::getEventMgr()->sendEvent(
      new EventCreateObject<FWorkflow,FWorkflow>(node, this) );
    myProc[id] = node;
    return node;
  }
  throw WfStructException(WfStructException::eDUPLICATE_NODE,"FNode id="+id);
}

FSourceNode*
FWorkflow::createSource(const string& id, WfCst::WfDataType type)
    throw (WfStructException)
{
  try {
    getInterfaceNode(id);
  } catch (WfStructException& e) {
    TRACE_TEXT (TRACE_ALL_STEPS,"Creating source node : " << id << endl);
    FSourceNode* node = new FSourceNode(this, id, type);
    myInterface[id] = node;
    return node;
  }
  throw WfStructException(WfStructException::eDUPLICATE_NODE,"FNode id="+id);
}

FConstantNode*
FWorkflow::createConstant(const string& id, WfCst::WfDataType type)
    throw (WfStructException)
{
  try {
    getInterfaceNode(id);
  } catch (WfStructException& e) {
    TRACE_TEXT (TRACE_ALL_STEPS,"Creating constant node : " << id << endl);
    FConstantNode* node = new FConstantNode(this, id, type);
    myInterface[id] = node;
    return node;
  }
  throw WfStructException(WfStructException::eDUPLICATE_NODE,"FNode id="+id);
}

FSinkNode*
FWorkflow::createSink(const string& id, WfCst::WfDataType type, unsigned int depth)
    throw (WfStructException)
{
  try {
    getInterfaceNode(id);
  } catch (WfStructException& e) {
    TRACE_TEXT (TRACE_ALL_STEPS,"Creating sink node : " << id << endl);
    FSinkNode* node = new FSinkNode(this, id, type, depth);
    myInterface[id] = node;
    return node;
  }
  throw WfStructException(WfStructException::eDUPLICATE_NODE,"FNode id="+id);
}

/**
 * Specific DFS search method
 * Skips the back nodes
 * NOT THREAD SAFE
 */

struct DFSNodeInfo {
  bool  explored;
  bool  ongoing;
  short end;
  DFSNodeInfo() {};
  DFSNodeInfo(bool _expl, bool _ongoing, short _end)
  : explored(_expl),ongoing(_ongoing),end(_end) {};
};

void DFS(WfNode* node,
         map<WfNode*, DFSNodeInfo>& DFSInfo,
         short& endCount) {
  DFSInfo[node].ongoing = true;
  TRACE_TEXT (TRACE_ALL_STEPS, "Starting DFS search for node : " << node->getId() << endl);
  for(list<WfNode*>::iterator nextIter = node->nextNodesBegin();
      nextIter != node->nextNodesEnd();
      ++nextIter) {
    WfNode* nextNode = (WfNode*) *nextIter;
    TRACE_TEXT (TRACE_ALL_STEPS, "  Processing next node : " << nextNode->getId() << endl);
    // look for next node DFS information
    map<WfNode*, DFSNodeInfo>::iterator nextNodeInfoSearch = DFSInfo.find(nextNode);
    // if not found skip the node (means next node is not a processor)
    if (nextNodeInfoSearch == DFSInfo.end())
      continue;
    DFSNodeInfo& nextNodeInfo = nextNodeInfoSearch->second;
    if (!nextNodeInfo.explored) {
      if (!nextNodeInfo.ongoing) {
        DFS(nextNode, DFSInfo, endCount);
      } else {
        // this is a BACK edge
        TRACE_TEXT (TRACE_ALL_STEPS, "Found back edge!!" << endl);
      }
    } else {
      TRACE_TEXT (TRACE_ALL_STEPS, "  Next node was already explored" << endl);
    }
  }
  DFSInfo[node].explored = true;
  DFSInfo[node].end      = endCount++;
  TRACE_TEXT (TRACE_ALL_STEPS, "End of DFS search for node : " << node->getId()
            << " (count=" << endCount-1 << ")" << endl);
}

/**
 * FWorkflow initialization
 * NOT THREAD SAFE
 * FIXME make it private ?
 */
void
FWorkflow::initialize() {

  if (getPortNb() > 0) {  // used for sub-workflows
    FProcNode::initialize();
  }

  TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "Sorting processors..." << endl);
  // create a topologically sorted list of processor nodes
  map<WfNode*, DFSNodeInfo>* DFSInfo = new map<WfNode*,DFSNodeInfo>();
  short DFSEndCount = 0;
  // -> initialization of DFSInfo
  for(map<string,FProcNode*>::iterator iter = myProc.begin();
      iter != myProc.end();
      ++iter) {
    WfNode* node = (WfNode*) iter->second;
    (*DFSInfo)[node] = DFSNodeInfo(false,false,0);
  }
  // -> DFS search
  for(map<string,FProcNode*>::iterator iter = myProc.begin();
      iter != myProc.end();
      ++iter) {
    WfNode* node = (WfNode*) iter->second;
    if (!(*DFSInfo)[node].explored) {
      DFS(node,*DFSInfo, DFSEndCount);
    }
  }
  // -> create sorted todo list
  for(map<string,FProcNode*>::iterator iter = myProc.begin();
      iter != myProc.end();
      ++iter) {
    FProcNode* node = (FProcNode*) iter->second;
    if (todoProc.empty()) {
      todoProc.push_back(node);
    } else { // insertion sort
      short insNodePrio = (*DFSInfo)[(WfNode*)node].end;
      list<FProcNode*>::iterator todoIter = todoProc.begin();
      while ((insNodePrio < (*DFSInfo)[(WfNode*) *todoIter].end)
              && (todoIter != todoProc.end()))
           ++todoIter;
      todoProc.insert(todoIter, node);
    }
  }

  TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "Connecting all nodes..." << endl);
  for(map<string,FProcNode*>::iterator iter = myProc.begin();
      iter != myProc.end();
      ++iter) {
    ((FProcNode*) iter->second)->connectNodePorts();
  }
  for(map<string,FNode*>::iterator iter = myInterface.begin();
      iter != myInterface.end();
      ++iter) {
    ((FNode*) iter->second)->connectNodePorts();
  }

  TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "Initializing Interface..." << endl);
  for(map<string,FNode*>::iterator iter = myInterface.begin();
      iter != myInterface.end();
      ++iter) {
    try {
      ((FNode*) iter->second)->initialize();
    } catch (WfDataException& e) {
        WARNING("Data failure during node " << ((FNode*) iter->second)->getId()
                << " initialization:" << endl << e.ErrorMsg() << endl);
    } catch (WfDataHandleException& e) {
        WARNING("Failure during node " << ((FNode*) iter->second)->getId()
                << " initialization:" << endl << e.ErrorMsg() << endl);
    }
  }
  TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "Initializing Processors..." << endl);
  for(map<string,FProcNode*>::iterator iter = myProc.begin();
      iter != myProc.end();
      ++iter) {
    ((FProcNode*) iter->second)->initialize();
  }
}

string
FWorkflow::toString() const {
  return "WORKFLOW " + getId();
}

/**
 * Set the path of the file used for data sources instanciation
 */
void
FWorkflow::setDataSrcXmlFile(const string& dataFileName) {
  this->dataSrcXmlFile = dataFileName;
}

/**
 * Returns the path of the file used for data sources instanciation
 */
const string&
FWorkflow::getDataSrcXmlFile() {
  return dataSrcXmlFile;
}

/**
 * FWorkflow instanciation
 * NOT THREAD SAFE
 */

void
FWorkflow::instanciate(Dag * dag) {
  // re-initialize status
  myStatus = N_INSTANC_READY;

  TRACE_TEXT (TRACE_MAIN_STEPS, traceId() << "########## WORKFLOW '" << getId()
                            << "' INSTANCIATION START ##########" << endl);

  if (getPortNb() > 0) {  // used for sub-workflows
    FProcNode::instanciate(dag); // run the iterator and call createReal/VoidInstance
  }
  TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "###### Instanciate sources/constants ..." << endl);
  // instanciate the INTERFACE (this will create the input data items if source is XML)
  for(map<string,FNode*>::iterator iter = myInterface.begin();
      iter != myInterface.end();
      ++iter)
  {
    // CONSTANT NODES
    FConstantNode* cstNode = dynamic_cast<FConstantNode*>((FNode*) iter->second);
    if (cstNode) {
      cstNode->instanciate(dag);
    }
    // SOURCE NODES
    FSourceNode* srcNode = dynamic_cast<FSourceNode*>((FNode*) iter->second);
    if (srcNode) {
      srcNode->resumeInstanciation();
      if (srcNode->instanciationReady()) {
        try {
          srcNode->instanciate(dag);
        } catch (WfDataHandleException& e) {
          WARNING("Failure during node " << srcNode->getId() << " instanciation:"
                  << endl << e.ErrorMsg());
        }
      }
      if (srcNode->instanciationCompleted())
        srcNode->finalize();
    }
  }

  TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "###### Instanciate processors..." << endl);
  // resume instanciation (if node was put on hold during a previous instanciate call)
  for(list<FProcNode*>::iterator iter = todoProc.begin();
      iter != todoProc.end();
      ++iter) {
     ((FProcNode*) *iter)->resumeInstanciation();
  }

  // loop until all nodes have no data to process
  bool dataToProcess = true;
  while ( dataToProcess ) {

    // for each node in the todo list
    list<FProcNode*>::iterator procIter = todoProc.begin();
    while (procIter != todoProc.end()) {
      FProcNode* currProc = (FProcNode*) *procIter;

      // instanciate node
      TRACE_TEXT (TRACE_MAIN_STEPS,"  ## INSTANCIATE NODE : " << currProc->getId() << endl);
      try {
        currProc->instanciate(dag);
      } catch (WfDataHandleException& e) {
        WARNING("Failure during node " << currProc->getId() << " instanciation:"
                << endl << e.ErrorMsg());
      } catch (WfDataException& e) {
        WARNING("Data failure during node " << currProc->getId() << " instanciation:"
                << endl << e.ErrorMsg());
      }
      // if on hold, then set the whole wf on hold
      if (currProc->instanciationOnHold()) {
        myStatus = N_INSTANC_ONHOLD;
      }
      if (currProc->instanciationPending()) {
        myStatus = N_INSTANC_PENDING;
      }
      // if fully instanciated, remove from todo list
      if (currProc->instanciationCompleted()) {
        currProc->finalize();
        TRACE_TEXT (TRACE_MAIN_STEPS,"#### Processor " << currProc->getId()
                      << " instanciation COMPLETE" << endl);
        procIter = todoProc.erase(procIter);
        continue;
      }
      ++procIter;
    } // end loop todo list

    TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "###### Check if still data to process..." << endl);
    // check if nodes have still some data to process
    // (this may happen in case of workflow restart with while loops when activities
    // inside the loop are re-used, ie they are already completed; then outputs of
    // these activities are submitted to the while node during the same round)
    dataToProcess = false;
    for (list<FProcNode*>::iterator procIter = todoProc.begin();
         procIter != todoProc.end();
         ++procIter) {
      FProcNode *proc = (FProcNode*) *procIter;
      if (proc->hasDataToProcess() && !proc->instanciationOnHold()) {
        dataToProcess = true;
        break;
      }
    }
    TRACE_TEXT (TRACE_ALL_STEPS, traceId() << (dataToProcess ? "YES" : "NO") << endl);

  } // end while (dataToProcess)

  TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "###### Instanciate sinks ..." << endl);
  for(map<string,FNode*>::iterator iter = myInterface.begin();
      iter != myInterface.end();
      ++iter) {
    FSinkNode* sinkNode = dynamic_cast<FSinkNode*>((FNode*) iter->second);
    if (sinkNode) {
      sinkNode->instanciate(dag);
      if (sinkNode->instanciationCompleted())
        sinkNode->finalize();
    }
  }

  if (instanciationOnHold()) {
    TRACE_TEXT (TRACE_MAIN_STEPS, traceId() << "########## WORKFLOW INSTANCIATION ON HOLD ##########" << endl);
  }
  // check if instanciation is pending on node execution
  // the first condition happens when one of the nodes (a sub-wf) is pending
  else if (instanciationPending() || !pendingNodes.empty()) {
    TRACE_TEXT (TRACE_MAIN_STEPS, traceId() << "########## WORKFLOW INSTANCIATION PENDING ##########" << endl);
    myStatus = N_INSTANC_PENDING;
  }
  // check if instanciation is finished
  else if (todoProc.empty()) {
    TRACE_TEXT (TRACE_MAIN_STEPS, traceId() << "############ WORKFLOW INSTANCIATION END ############" << endl);
    // check if sinks are completed too
    for(map<string,FNode*>::iterator iter = myInterface.begin();
      iter != myInterface.end();
      ++iter) {
      FSinkNode* sinkNode = dynamic_cast<FSinkNode*>((FNode*) iter->second);
      if (sinkNode && !sinkNode->instanciationCompleted()) {
        WARNING(traceId() << "Sink '" << sinkNode->getId() << " is not complete (instanciation error)" << endl);
      }
    }
    myStatus = N_INSTANC_END;
  }
  else if (getPortNb() > 0) {  // used for sub-workflows
    TRACE_TEXT (TRACE_MAIN_STEPS, traceId() << "############ WORKFLOW INSTANCIATION NOT FINISHED ############" << endl);
  }
  // incorrect status
  else {
    WARNING(traceId() << "Instanciation stopped in incorrect state!!");
    myStatus = N_INSTANC_STOPPED;
    for (list<FProcNode*>::iterator procIter = todoProc.begin();
         procIter != todoProc.end();
         ++procIter) {
      WARNING(traceId() << "Functional node remaining in TODO list: "
                  << ((FProcNode*) *procIter)->getId() << endl);
    }
  }
  // add dag to my list of dags
  myDags.push_back(dag);
}

/**
 * Create one instance of the workflow (USED FOR SUB-WORKFLOWS)
 */
void
FWorkflow::createRealInstance(Dag* dag,
                              const FDataTag& currTag,
                              vector<FDataHandle*>& currDataLine) {
  // LOOP for each source
  for (map<string, FNode *>::iterator interfIter = myInterface.begin();
       interfIter != myInterface.end();
       ++interfIter) {
    FSourceNode * source = dynamic_cast<FSourceNode*>((FNode*) interfIter->second);
    if (source) {
      source->createInstance(currTag, currDataLine);
    }
  } // end for interface nodes
}

/**
 * Create one VOID instance of the workflow (USED FOR SUB-WORKFLOWS)
 */
void
FWorkflow::createVoidInstance(const FDataTag& currTag,
                              vector<FDataHandle*>& currDataLine) {
  createRealInstance(NULL,currTag,currDataLine);
}

/**
 * Update instanciation status: USED FOR SUB-WORKFLOWS ONLY
 */
void
FWorkflow::updateInstanciationStatus() {
  if (myRootIterator->isAtEnd()) {
    if (myRootIterator->isDone()) {

      // LOOP for each source - set them as completed
      // (sources cannot update their status themselves as they don't use an iterator)
      for (map<string, FNode *>::iterator interfIter = myInterface.begin();
           interfIter != myInterface.end();
           ++interfIter) {
        FSourceNode * source = dynamic_cast<FSourceNode*>((FNode*) interfIter->second);
        if (source) {
          source->setInstanciationCompleted();
        }
      } // end for interface nodes

      TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "########## ALL INPUTS PROCESSED" << endl);
      // status not updated here but in the instanciate() method because instanciation
      // is completed only after all nodes of the wf have been completed

    } else {
      TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "########## WAITING FOR INPUTS " << endl);
      myStatus = N_INSTANC_READY;
    }
  }
}

void
FWorkflow::downloadSinkData() {
  // display all sink results
  for (map<string,FNode*>::const_iterator nodeIter = myInterface.begin();
       nodeIter != myInterface.end();
       ++nodeIter) {
    FSinkNode *sink = dynamic_cast<FSinkNode*>((FNode*) nodeIter->second);
    if (sink != NULL) {
      sink->downloadResults();
    }
  }
}

void
FWorkflow::displayAllResults(ostream& output) {
  // display cancelled dags
  for (list<Dag*>::const_iterator dagIter = myDags.begin();
       dagIter != myDags.end();
       ++dagIter) {
    Dag * currDag = (Dag*) *dagIter;
    if (currDag->isCancelled()) {
       output  << "** DAG " << currDag->getId()
               << " was cancelled => no results **" << endl;
    }
  }
  // display all sink results
  for (map<string,FNode*>::const_iterator nodeIter = myInterface.begin();
       nodeIter != myInterface.end();
       ++nodeIter) {
    FSinkNode *sink = dynamic_cast<FSinkNode*>((FNode*) nodeIter->second);
    if (sink != NULL) {
      sink->displayResults(output);
    }
  }
}

void
FWorkflow::getSinkContainer(const string& sinkName,
                            string& containerID) {
  FNode* interfNode = getInterfaceNode(sinkName);
  FSinkNode *sink = dynamic_cast<FSinkNode*>(interfNode);
  if (sink)
    try {
      sink->getResultsInContainer(containerID);
    } catch (WfDataHandleException& e) {
      WARNING("Cannot get sink container ID (" << e.ErrorMsg() << ")" << endl);
    }
  else
    throw WfStructException(WfStructException::eUNKNOWN_NODE,
                            "Sink name="+sinkName);
}

void
FWorkflow::displayDagSummary(ostream& output) {
  output << "** INSTANCIATED DAGS SUMMARY:" << endl;
  for (list<Dag*>::iterator dagIter = myDags.begin();
       dagIter != myDags.end();
       ++dagIter) {
    Dag * dag = (Dag*) *dagIter;
    output << "DAG ID '" << dag->getId() << "': "
           << "size = " << dag->size() << " node(s), "
           << "makespan = " << dag->getMakespan() << " ms" << endl;
  }
}

void
FWorkflow::setPendingInstanceInfo(DagNode * dagNode,
                                  FDataHandle * dataHdl,
                                  FNodeOutPort * outPort,
                                  FNodeInPort * inPort) {
  pendingDagNodeInfo_t info;
  info.dataHdl = dataHdl;
  info.outPort = outPort;
  info.inPort  = inPort;
  pendingNodes.insert(make_pair(dagNode,info));
  TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "new PENDING dag node ("
                              << dagNode->getId() << ") for IN port (" << inPort->getCompleteId()
                              << ")" << endl);
}

/**
 * Handle end of node execution
 */
void
FWorkflow::handlerDagNodeDone(DagNode* dagNode) {
  bool statusChange = false;
  myLock.lock();    /** LOCK */
  // search the instance in the pending list
  multimap<DagNode*, pendingDagNodeInfo_t>::iterator pendingIter, pendingStart;
  pendingStart = pendingIter = pendingNodes.lower_bound(dagNode);
  // if found, resubmit the datahandle to the in port(s)
  while (pendingIter != pendingNodes.upper_bound(dagNode)) {
    pendingDagNodeInfo_t info = (pendingDagNodeInfo_t) (pendingIter++)->second;
    TRACE_TEXT (TRACE_ALL_STEPS, traceId() << " ## RETRY DATA SUBMISSION FOR INSTANCE : "
                                << dagNode->getId() << endl);

    try {
      // update the data ID
      info.dataHdl->downloadDataID();
      // submit data to the in port
      info.outPort->reSendData(info.dataHdl, info.inPort);

    } catch (WfDataHandleException& e) {
      WARNING("Instanciator error due to data handle error :" << e.ErrorMsg());
      myStatus = N_INSTANC_STOPPED;
    } catch (WfDataException& e) {
      WARNING("Instanciator error due to invalid dag data :" << e.ErrorMsg());
      myStatus = N_INSTANC_STOPPED;
    }
    // re-start instanciation
    statusChange = true;

    TRACE_TEXT (TRACE_ALL_STEPS, traceId() << " ## END OF RETRY FOR INSTANCE : "
                                << dagNode->getId() << endl);
  }
  // remove entries from the pending list
  pendingNodes.erase(pendingStart,pendingIter);
  // re-initialize instanciation
  if (statusChange)
    resumeInstanciation();

  myLock.unlock();  /** UNLOCK */
}

/**
 * Status
 */
void
FWorkflow::writeAllDagsState(ostream& output) {
  TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "Writing workflow status ..." <<  endl);
  output << "<dags>" << endl;
  for (list<Dag*>::iterator dagIter = myDags.begin();
       dagIter != myDags.end();
       ++dagIter) {
    Dag *currDag = ((Dag*) *dagIter);
    if (!currDag->size() == 0) {
      currDag->toXML(output);
    }
  }
  output << "</dags>" << endl;
}

void
FWorkflow::writeAllSourcesState(ostream& output) {
  // my own sources
  for(map<string,FNode*>::iterator iter = myInterface.begin();
      iter != myInterface.end();
      ++iter) {
    FSourceNode* srcNode = dynamic_cast<FSourceNode*>((FNode*) iter->second);
    if (srcNode && !srcNode->isConnectedToWfPort())
      srcNode->toXML(output);
  }
  // idem for all sub-workflows
  for(map<string,FProcNode*>::iterator iter = myProc.begin();
      iter != myProc.end();
      ++iter) {
    FWorkflow* subWf = dynamic_cast<FWorkflow*>((FProcNode*) iter->second);
    if (subWf)
      subWf->writeAllSourcesState(output);
  }
}

void
FWorkflow::writeAllSinksState(ostream& output) {
  // my own sinks
  for(map<string,FNode*>::iterator iter = myInterface.begin();
      iter != myInterface.end();
      ++iter) {
    FSinkNode* sinkNode = dynamic_cast<FSinkNode*>((FNode*) iter->second);
    if (sinkNode && !sinkNode->isConnectedToWfPort())
      sinkNode->toXML(output);
  }
}

void
FWorkflow::writeAllSourcesAndSinksData(ostream& output) {
  output << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << endl;
  output << "<data>" << endl;
  this->writeAllSourcesState(output);
  this->writeAllSinksState(output);
  output << "</data>" << endl;
}

/**
 * Restart workflow
 * This method stores a ptr to the dag nodes that are already done in the
 * transcriptNodes map, and include the transcript dags (if not empty) in the workflow
 * dag list
 */
void
FWorkflow::readDagsState(list<Dag*>& dagList) {
  TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "Using previous execution transcript ..." <<  endl);
  while (!dagList.empty()) {
    Dag *currDag = dagList.front();
    if (currDag) {
      // iterate over all nodes
      //	if done => store (node ID ====> dagNode ptr)
      //        if not done => remove from dag
      for ( map<string, DagNode *>::iterator nodeIter = currDag->begin();
            nodeIter != currDag->end();
            ++nodeIter ) {
	DagNode * currNode = (DagNode*) nodeIter->second;
	if (currNode->isDone()) {
	  transcriptNodes[currNode->getId()] = currNode;
	} else {
          currDag->removeNode(currNode->getId());
          delete currNode;
        }
      }
      // store the cleaned dag in current workflow's dag list
      if (!currDag->size() == 0)
        myDags.push_back(currDag);
      else
        delete currDag;
    }
    dagList.pop_front();
  }
  TRACE_TEXT (TRACE_ALL_STEPS, traceId() << " ==> " << transcriptNodes.size()
					 << " executed nodes found" << endl);
}

void
FWorkflow::findDagNodeTranscript( const string& dagNodeId,
                                  DagNode* & dagNodePtr,
                                  bool& isDone ) {
  map<string, DagNode*>::iterator nodeIter = transcriptNodes.find(dagNodeId);
  if (nodeIter != transcriptNodes.end()) {
    // node found in current workflow transcript map
    dagNodePtr = (DagNode*) nodeIter->second;
    isDone = dagNodePtr->isDone();
  } else if (getWorkflow() != NULL) {
    // ask parent workflow (if sub-workflow)
    getWorkflow()->findDagNodeTranscript(dagNodeId, dagNodePtr, isDone);
  } else {
    // node not found
    dagNodePtr = NULL;
    isDone = false;
  }
}

/**
 * Memory free
 */

void
FWorkflow::deleteAllResults() {
  TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "Deleting all dag results ..." <<  endl);
  for (list<Dag*>::iterator dagIter = myDags.begin();
       dagIter != myDags.end();
       ++dagIter) {
    ((Dag*) *dagIter)->deleteAllResults();
  }
}

void
FWorkflow::deleteAllInputData(MasterAgent_var& MA) {
  TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "Deleting all wf input data..." <<  endl);
  for(map<string,FNode*>::iterator iter = myInterface.begin();
      iter != myInterface.end();
      ++iter) {
    ((FNode*) iter->second)->freeNodePersistentData(MA);
  }
}

void
FWorkflow::deleteAllIntermediateData(MasterAgent_var& MA) {
  TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "Deleting all wf intermediate data..." <<  endl);
  for(map<string,FProcNode*>::iterator iter = myProc.begin();
      iter != myProc.end();
      ++iter) {
    ((FNode*) iter->second)->freeNodePersistentData(MA);
  }
}

void
FWorkflow::freeNodePersistentData(MasterAgent_var& MA) {
  TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "Deleting sub-workflow data..." <<  endl);
  deleteAllInputData(MA);
  deleteAllIntermediateData(MA);
}

void
FWorkflow::deleteAllDags() {
  TRACE_TEXT (TRACE_ALL_STEPS, traceId() << "Deleting all wf dags ..." <<  endl);
  while (! myDags.empty() ) {
    Dag * p = myDags.front();
    myDags.pop_front();
    delete p;
  }
}
