/****************************************************************************/
/* The class representing a functional workflow                             */
/* Contains the workflow nodes and manages their instanciation as one or    */
/* several dags.                                                            */
/*                                                                          */
/* Author(s):                                                               */
/* - Benjamin ISNARD (benjamin.isnard@ens-lyon.fr)                          */
/*                                                                          */
/*  This file is part of DIET .                                             */
/*                                                                          */
/*  Copyright ENS Lyon, INRIA, UCBL, SysFera (2000)                         */
/*                                                                          */
/*  - Frederic.Desprez@ens-lyon.fr (Project Manager)                        */
/*  - Eddy.Caron@ens-lyon.fr (Technical Manager)                            */
/*  - Tech@sysfera.com (Maintainer and Technical Support)                   */
/*                                                                          */
/*  This software is a computer program whose purpose is to provide an      */
/*  easy and transparent access to distributed and heterogeneous            */
/*  platforms.                                                              */
/*                                                                          */
/*                                                                          */
/*  This software is governed by the CeCILL license under French law and    */
/*  abiding by the rules of distribution of free software.  You can  use,   */
/*  modify and/ or redistribute the software under the terms of the CeCILL  */
/*  license as circulated by CEA, CNRS and INRIA at the following URL       */
/*  "http://www.cecill.info".                                               */
/*                                                                          */
/*  As a counterpart to the access to the source code and  rights to copy,  */
/*  modify and redistribute granted by the license, users are provided      */
/*  only with a limited warranty  and the software's author,  the holder    */
/*  of the economic rights,  and the successive licensors  have only        */
/*  limited liability.                                                      */
/*                                                                          */
/*  In this respect, the user's attention is drawn to the risks             */
/*  associated with loading,  using,  modifying and/or developing or        */
/*  reproducing the software by the user in light of its specific status    */
/*  of free software, that may mean  that it is complicated to              */
/*  manipulate, and  that  also therefore means  that it is reserved for    */
/*  developers and experienced professionals having in-depth computer       */
/*  knowledge. Users are therefore encouraged to load and test the          */
/*  software's suitability as regards their requirements in conditions      */
/*  enabling the security of their systems and/or data to be ensured and,   */
/*  more generally, to use and operate it in the same conditions as         */
/*  regards security.                                                       */
/*                                                                          */
/*  The fact that you are presently reading this means that you have had    */
/*  knowledge of the CeCILL license and that you accept its terms.          */
/*                                                                          */
/****************************************************************************/
/* $Id$
 * $Log$
 * Revision 1.20  2010/07/20 09:20:11  bisnard
 * integration with eclipse gui and with dietForwarder
 *
 * Revision 1.19  2009/09/25 12:39:18  bisnard
 * modified includes to reduce inter-dependencies
 *
 * Revision 1.18  2009/08/26 10:33:09  bisnard
 * implementation of workflow status & restart
 *
 * Revision 1.17  2009/07/30 09:43:09  bisnard
 * modified methods to free persistent data generated by workflow execution
 *
 * Revision 1.16  2009/07/23 12:30:10  bisnard
 * new method finalize() for functional wf nodes
 * removed const on currDataLine parameter for instance creation
 *
 * Revision 1.15  2009/07/10 12:55:59  bisnard
 * implemented while loop workflow node
 *
 * Revision 1.14  2009/07/07 09:03:22  bisnard
 * changes for sub-workflows (FWorkflow class now inherits from FProcNode)
 *
 * Revision 1.13  2009/06/23 12:23:58  bisnard
 * removed dependencies (includes)
 *
 * Revision 1.12  2009/05/27 08:49:43  bisnard
 * - modified condition output: new IF_THEN and IF_ELSE port types
 * - implemented MERGE and FILTER workflow nodes
 *
 * Revision 1.11  2009/05/15 11:10:20  bisnard
 * release for workflow conditional structure (if)
 *
 * Revision 1.10  2009/04/17 08:54:43  bisnard
 * renamed Node class as WfNode
 *
 * Revision 1.9  2009/04/09 09:56:20  bisnard
 * refactoring due to new class FActivityNode
 *
 * Revision 1.8  2009/04/08 09:34:56  bisnard
 * pending nodes mgmt moved to FWorkflow class
 * FWorkflow and FNode state graph revisited
 * FNodePort instanciation refactoring
 *
 * Revision 1.7  2009/02/06 14:55:08  bisnard
 * setup exceptions
 *
 * Revision 1.6  2009/01/16 16:31:54  bisnard
 * added option to specify data source file name
 *
 * Revision 1.5  2009/01/16 13:54:50  bisnard
 * new version of the dag instanciator with iteration strategies for nodes with multiple input ports
 *
 * Revision 1.4  2008/12/09 12:15:59  bisnard
 * pending instanciation handling (uses dag outputs for instanciation
 * of a functional wf)
 *
 * Revision 1.3  2008/12/02 10:09:36  bisnard
 * added instanciation methods
 *
 * Revision 1.2  2008/10/22 09:29:00  bisnard
 * replaced uint by standard type
 *
 * Revision 1.1  2008/10/20 08:09:58  bisnard
 * new classes for functional workflows
 *
 */

#ifndef _FWORKFLOW_HH_
#define _FWORKFLOW_HH_

#include <map>
#include <iostream>

#include "NodeSet.hh"
#include "FNode.hh"

using namespace std;

/* The different types of nodes used in a functional workflow */

class FActivityNode;
class FIfNode;
class FMergeNode;
class FFilterNode;
class FLoopNode;

class FWorkflow : public NodeSet, public FProcNode {

public:

  /***************************************************/
  /* constructors/destructor                         */
  /***************************************************/

  /**
   * Constructor of toplevel workflow (no parent)
   * @param id	  workflow identifier
   * @param name  name of the workflow (only descriptive)
   */
  FWorkflow(string id, string name);
  
  /**
   * Constructor of sub-workflow (child of another workflow)
   * @param id	      a unique identifier (among all parent wf processors)
   * @param parentWf  ref to the parent wf
   */
  FWorkflow(string id, string name, FWorkflow* parentWf);

  /**
   * FWorkflow destructor
   * Note: it does not delete the instantiated dags (use deleteAllDags)
   */
  virtual ~FWorkflow();
  
  /***************************************************/
  /* event message types                             */
  /***************************************************/
  
  enum eventMsg_e {
    INST,
    INSTDONE,
    INSTERROR,
    COMPLETED,
    CANCELLED
  };

  /***************************************************/
  /*            NodeSet methods                      */
  /***************************************************/

  virtual WfNode *
  getNode(const string& nodeId) throw (WfStructException);

  virtual void
  checkPrec(NodeSet* contextNodeSet) throw (WfStructException);

  /***************************************************/
  /*               FNode methods                     */
  /***************************************************/

  virtual void
      initialize();

  virtual void
      instanciate(Dag* dag);

  virtual void
      freeNodePersistentData(MasterAgent_var& MA);

  /***************************************************/
  /*             FProcNode methods                   */
  /***************************************************/
  virtual void
      createRealInstance(Dag* dag,
                         const FDataTag& currTag,
                         vector<FDataHandle*>& currDataLine);

  virtual void
      createVoidInstance(const FDataTag& currTag,
                         vector<FDataHandle*>& currDataLine);

  virtual void
      updateInstanciationStatus();

  /***************************************************/
  /*               public methods                    */
  /***************************************************/
  string
  getName() const;
  
  virtual string
  toString() const;
  
  virtual FWorkflow*
  getRootWorkflow() const;

  void
  setDataSrcXmlFile(const string& dataFileName);

  const string&
  getDataSrcXmlFile();

  /**
   * Factory methods for workflow nodes
   * @return pointer to the node (does not return NULL)
   */

  FActivityNode*
  createActivity(const string& id) throw (WfStructException);

  FIfNode*
  createIf(const string& id) throw (WfStructException);

  FMergeNode*
  createMerge(const string& id) throw (WfStructException);

  FFilterNode*
  createFilter(const string& id) throw (WfStructException);

  FLoopNode*
  createLoop(const string& id) throw (WfStructException);

  FWorkflow*
  createSubWorkflow(const string& id, const string& name) throw (WfStructException);

  FSourceNode*
  createSource(const string& id, WfCst::WfDataType type) throw (WfStructException);

  FConstantNode*
  createConstant(const string& id, WfCst::WfDataType type) throw (WfStructException);

  FSinkNode*
  createSink(const string& id, WfCst::WfDataType type, unsigned int depth)
      throw (WfStructException);

  /**
   * Get methods for workflow nodes
   * @return pointer to the node (does not return NULL)
   */

  FProcNode *
  getProcNode(const string& id) throw (WfStructException);

  FNode *
  getInterfaceNode(const string& id) throw (WfStructException);


  /**
   * Setup the information used when node instance (dagnode) is done
   * and some output ports instances have to be re-submitted to next
   * nodes (when cardinal was not known before node execution)
   * @param dagNode the dag node
   * @param dataHdl the data handle that could not be sent to the in port
   * @param outPort the FNode out port producing the data
   * @param inPort  the FNode in port receiving the data
   */
  void
  setPendingInstanceInfo(DagNode * dagNode,
                         FDataHandle * dataHdl,
                         FNodeOutPort * outPort,
                         FNodeInPort * inPort);

  /**
   * Handles the dagNode end of execution event
   * Uses the info set with setPendingInstanceInfo to re-send the
   * data to connected nodes
   * @param dagNode the dag node
   */
  void
  handlerDagNodeDone(DagNode* dagNode);

  /**
   * Write a complete status of the workflow on a stream (XML encoded)
   * This status contains the current state of all generated dags
   * @param output  output stream
   */
  void
  writeAllDagsState(ostream& output);

  /**
   * Write a status of all the sources and sinks of the workflow
   * including sub-workflows sources. This can be used as input for
   * another execution of the same data to avoid re-uploading all
   * source data to the platform. (uses existing data IDs)
   * @param output  output stream
   */
  void
  writeAllSourcesAndSinksData(ostream& output);

  /**
   * Initialize a workflow using the transcript of all generated
   * dags from a previous execution of the workflow. Only
   * tasks that have not been executed during that execution will
   * be instanciated and executed. Data already saved on the platform
   * will be re-used (uses data IDs written in the transcript)
   */
  void
  readDagsState( list<Dag*>& dagList );

  /**
   * Check workflow execution transcript for a given dag node id
   * @param dagNodeId	id of the node (must match)
   * @param dagNodePtr	returns a ptr to the dagNode if found
   * @param isDone	returns true if dagNode is done
   */
  void findDagNodeTranscript( const string& dagNodeId,
			      DagNode* & dagNodePtr,
			      bool& isDone );

  /**
   * results
   */

  /**
   * Download all sinks data description
   * (data IDs for all nodes and leaves + values for leaves only)
   * Must be called before other results methods
   */
  void
  downloadSinkData();

  void
  displayAllResults(ostream& output);
  void
  displayDagSummary(ostream& output);
  void
  getSinkContainer(const string& sinkName,
                   string& containerID);

  /**
   * memory free
   */

  /**
   * Remove all data produced by the workflow from the platform
   */
  void deleteAllResults();

  /**
   * Remove input data created by the workflow engine on the platform
   * (sources and constants)
   */
  void deleteAllInputData(MasterAgent_var& MA);

  /**
   * Remove intermediate data created by the workflow engine on the platform
   */
  void deleteAllIntermediateData(MasterAgent_var& MA);

  /**
   * Free memory used for dags generated by the workflow engine
   */
  void deleteAllDags();


private:
  
  /**
   * Workflow Name
   * (descriptive)
   */
  string myName;

  /**
   * Workflow nodes for the interface (sources, sinks, constants)
   */
  map<string, FNode *> myInterface;

  /**
   * Workflow nodes for the processors
   */
  map<string, FProcNode*> myProc;

  /**
   * Workflow nodes to be instanciated (ordered list)
   */
  list<FProcNode*> todoProc;

  /**
   * File to use for instanciation of data sources
   */
  string dataSrcXmlFile;

  /**
   * List of generated dags
   */

  list<Dag*> myDags;

  /**
   * Pending instance record
   */
  struct pendingDagNodeInfo_t {
    FDataHandle * dataHdl;
    FNodeOutPort * outPort;
    FNodeInPort * inPort;
  };

  /**
   * The list of instances that will trigger new instanciation (due to
   * dependency of other functional nodes on the cardinal of the outputs of
   * this node)
   */
  multimap<DagNode*, pendingDagNodeInfo_t> pendingNodes;

  /**
   * The execution transcript of the workflow
   * It contains all dagNodes that have been executed (status isDone)
   */
  map<string, DagNode*>	transcriptNodes;

  /**
   * Critical section
   */
  omni_mutex myLock;

  /***********************/
  /*   PRIVATE METHODS   */
  /***********************/

  void writeAllSourcesState(ostream& output);
  void writeAllSinksState(ostream& output);

};


#endif // _FWORKFLOW_HH_




